; New equates
PROJ_UP		=	1
PROJ_LEFT	=	2
PROJ_RIGHT	=	4
PROJ_DOWN	=	8

; External procedures for 1STCELL

safe_putch:
; Safely place a character onto the screen to avoid snow.
; Call with:
;	AX = character attributes
; Preserves: AX, BX, DX
; Destroys: DI (advanced by 2, can still be managed)

; If building for MDA, the bit checking instructions will not be assembled,
; and this procedure will only execute stosw.
	SAFE_PUTCH
	ret

setpos:
; Assign a new destination index based on the X and Y position of something.
; Call with:
;	DX = X and Y positions
; Preserves: AX, CX
; Returns: DI
	push		ax
	xor		ah,ah
	mov		al,dh
	; (y << 5) + (y << 3) + x
	; (for 80x25, increment both shifts by 1)
	shl		ax,1				; New approach looks to be faster on 8088,
	shl		ax,1				; even if it is disgusting
	shl		ax,1
	SHLEXTRA
	push		cx
	mov		cx,ax				; Point where (y << 3) took effect, save that now
	shl		ax,1
	shl		ax,1
	add		ax,cx				; Add (y << 5) and (y << 3) together now
	xor		ch,ch

	mov		cl,dl
	add		ax,cx				; Add X value to index
	shl		ax,1				; Double it to give a valid DI value

	pop		cx
	mov		di,ax
	pop		ax

	ret

adjpos:
; Adjust the destination index based on old and new positions.
; This should be called AFTER setpos when DI is already set, and assumes
; a half of DX has been incremented or decremented once.
; It's faster than calling setpos a second time.
; Call with:
;	AX = old location
;	DX = new location
; Returns: DI
	sub		di,2			; Restore previous position of DI
	cmp		dl,al
	jg		adjright
	je		adjup
	sub		di,2
	jmp		adjup
adjright:
	add		di,2
adjup:
	cmp		dh,ah
	jg		adjdown
	je		adjend
	sub		di,X_MAX*2
	jmp		adjend
adjdown:
	add		di,X_MAX*2

adjend:
	ret

moveplayer:
; Move the player according to the keys being held, if any.
	
	; Only on odd ticks (i.e. once every two ticks...)
	test		[gtimer],1
	jz		stopmove
	
	mov		bl,[moveshootkeys]
	test		bl,0Fh
	jz		stopmove
	
	; Clear character at current position
	mov		dx,word [player_x]
	call		setpos
	xor		ax,ax
	call		safe_putch
	mov		ax,dx
	call		moveobject	

plcollide:
	call		collidewall
	jz		ploldloc
	mov		word [player_x],dx
	jmp		dispplch
ploldloc:
	mov		dx,ax			; Restore old destination index in case of collision

dispplch:
	call		adjpos
	mov		ax,(15*256)+PLAYER_CH
	call		safe_putch
	; (5/19/2022) I cut out the movement sound.

	; Check if the player was dumb enough to run into someone
	mov		si,eattr+2-OBJ_LEN
	mov		cx,ENEMY_CT
	call		checkcollide
	jnz		plwormcollide
	or		[gflags],1

plwormcollide:
	call		wormcollide
	jnz		stopmove
	or		[gflags],1

stopmove:
	ret

playershoot:
; The player shoots a bullet in a direction if WASD keys are held.
; Destroys: AX, CX, DX, SI
	mov		cl,[moveshootkeys]
	test		cl,0F0h
	jz		stopshoot
	mov		dx,word [player_x]
	xor		al,al
	xor		bl,bl
	
; Note: Should know how to handle diagonal fire...
pshup:
	test		cl,HKEY_W
	jz		pshdown
	mov		bl,PROJ_UP
	mov		al,'|'
	jmp		pshleft
pshdown:
	test		cl,HKEY_S
	jz		pshleft
	mov		bl,PROJ_DOWN
	mov		al,'|'
pshleft:
	test		cl,HKEY_A
	jz		pshright
	or		bl,PROJ_LEFT
	test		bl,PROJ_UP			; northwest?
	jz		pshsw
	mov		al,'\'
	jmp		pshok
pshsw:
	test		bl,PROJ_DOWN
	jz		pshhorz
	mov		al,'/'
	jmp		pshok

pshright:
	test		cl,HKEY_D
	jz		pshhorz
	or		bl,PROJ_RIGHT
	test		bl,PROJ_UP			; northeast?
	jz		pshse
	mov		al,'/'
	jmp		pshok
pshse:
	test		bl,PROJ_DOWN
	jz		pshhorz
	mov		al,'\'
	jmp		pshok
	
pshhorz:
	cmp		al,'|'
	je		pshok
	mov		al,'-'
	
pshok:
	mov		ah,0Fh
	mov		si,pbullet-OBJ_LEN
	mov		cx,PBULLET_CT				; four objects
	xor		bh,bh			; always takes one tick to move
	call		createprojectile
	jnz		stopshoot		; ZF clear = failed to create projectile
	or		[sflags],SSHOOT
	
stopshoot:
	ret

moveprojectiles:
; Every dumb projectile in an address is updated.
; Call with:
;	CX = count of words at buffer address
;	SI = address of projectile data
	
	; Does it exist?
	cmp		byte [si],0
	jnz		checkprojticks
	jmp		mvpjrepeat
	
checkprojticks:
	; Ready to move?
	mov		al,byte [si+5]
	call		checkticks
	je		processprojectile
	inc		byte [si+5]
	jmp		mvpjrepeat

processprojectile:
	; Reset projectile ticks
	and		byte [si+5],0F0h
	mov		dx,word [si+2]			; load location data
	call		setpos
	
	; Do not clear the original position if a player, enemy, or worm object exists there
	cmp		word [player_x],dx
	je		projsetdi

	mov		ax,si
	mov		bx,cx
	mov		cx,ENEMY_CT
	mov		si,eattr+2-OBJ_LEN
	call		checkcollide
	mov		cx,bx
	mov		si,ax
	jz		projsetdi
	call		wormcollide
	jz		projsetdi
	
	; Clear projectile's original position
	xor		ax,ax
	call		safe_putch
	sub		di,2				; sub takes less clocks than jmp

projsetdi:
	add		di,2
startprojmove:
	mov		ax,dx
	mov		bl,[si+4]			; load direction
	call		moveobject
setproj:
	; Did this projectile hit a wall?
	call		collidewall
	jnz		projbombcheck
	cmp		byte [si+6],PROJ_BOUNCE		; Can it bounce?
	je		projbounce
	jmp		destroyprojectile
projbounce:
	test		bl,0F0h
	jnz		projbounce2
	jmp		destroyprojectile
projbounce2:
	sub		bl,10h				; That's one less bounce for it
	call		bounce
	mov		byte [si+4],bl
	mov		dx,ax

projbombcheck:
	; Is this the player's bomb? If so, it should destroy other projectiles.
	cmp		si,pbomb
	jb		setproj2
	cmp		si,eattr
	jae		setproj2
	push		ax
	mov		ax,si
	mov		bx,cx				; may need to change to push...?
	mov		cx,PROJ_CT
	mov		si,eproj+2-OBJ_LEN
	call		checkcollide
	jnz		setproj1
	mov		word [si-2],0
	mov		word [si],0
setproj1:
	mov		cx,bx
	mov		si,ax
	pop		ax

	; Did a projectile hit the player?
	; (exclude the player's own projectiles)
setproj2:
	cmp		si,eattr
	jb		projputch
	cmp		dx,word [player_x]
	jne		projputch
	or		[gflags],1
	jmp		stopmoveproj			; If hit, processing other projectiles is a waste

projputch:
	; Create MORE projectiles if this bullet ripples!!
	; Now update the projectile character on the screen...
	call		adjpos
	mov		[si+2],dx

projtype:
	; Create more projectiles if this one ripples or is a laser
	test		bl,0F0h				; Do not attempt if the auxiliary halfbyte is zero
	jz		projputch2
	mov		bh,byte [si+6]
	cmp		bh,PROJ_RIPPLE			; Rippling projectiles?
	je		projrippler
	cmp		bh,PROJ_LASER			; A big long laser?
	je		projlaser
	jmp		projputch2

projlaser:
	; Laser handler
	mov		bh,byte [si+5]
	sub		bl,10h
	mov		dx,ax				; New laser object takes old position
	push		si
	push		cx
	mov		ax,[si]
	mov		si,eproj-OBJ_LEN
	mov		cx,PROJ_CT
	call		createprojectile
	mov		byte [si+6],PROJ_LASER		; Pass the laser byte over to the next one
	pop		cx
	pop		si
	mov		byte [si+6],0			; Old one cannot be a laser projectile anymore
	jmp		projputch3
projrippler:
	sub		byte [si+4],10h
	and		bl,0Fh
	mov		bh,byte [si+5]
	push		si
	push		cx
	mov		ax,[si]
	mov		si,eproj
	mov		cx,PROJ_CT
	call		ripple
	pop		cx
	pop		si
projputch2:
	mov		ax,[si]				; load character
projputch3:
	; Do not attempt to overwrite worm body parts; they are indestructible, remember?
	call		wormcollide
	jz		mvpjrepeat
	call		safe_putch
	jmp		mvpjrepeat

destroyprojectile:
	mov		word [si],0
	mov		word [si+2],0
	mov		byte [si+6],0
	; Rest should be left with garbage data

mvpjrepeat:
	; Move to next object in the address
	add		si,OBJ_LEN
	dec		cx
	jz		stopmoveproj
	jmp		moveprojectiles
	;loop		moveprojectiles
stopmoveproj:
	ret

collidewall:
; Check if an object collides with the wall.
; Call with:
; 	DX = location of current object
; Returns: ZF (if it hit the wall)
	cmp		dl,X_LEFT
	jle		cwend
	cmp		dl,X_RIGHT-1
	jge		cwend
	cmp		dh,1
	jle		cwend
	cmp		dh,Y_MAX-1
	jge		cwend

cwend:
	ret
	
moveobject:
; Move an object based on a direction being held.
; Call with:
;	BL = directional bits
;	DX = X and Y positions
; Returns: DX
	test		bl,PROJ_UP
	jz		objdown
	dec		dh
	jmp		objleft
objdown:
	test		bl,PROJ_DOWN
	jz		objleft
	inc		dh
objleft:
	test		bl,PROJ_LEFT
	jz		objright
	dec		dl
	jmp		mvobjend
objright:
	test		bl,PROJ_RIGHT
	jz		mvobjend
	inc		dl
mvobjend:
	ret

checkcollide:
; Check if an object collides with another.
; Call with:
;	CX = number of objects to analyze in the address
;	DX = location of current object
;	SI = address of objects to test against (position at location data! try appending +2-OBJ_LEN)
; Preserves: AX
; Returns: ZF (if they collide)

	; First, evaluate if the object hit the wall.	
ccnextpos:
	add		si,OBJ_LEN
	cmp		dx,[si]
ccskip:
	loopne		ccnextpos

ccend:
	ret

bounce:
; Bounce a projectile when it hits something.
; Call with:
;	BL = direction
;	DX = location
; Returns: BL
	cmp		dh,1
	jg		bounces
	and		bl,not PROJ_UP
	or		bl,PROJ_DOWN
	jmp		bouncee
bounces:
	cmp		dh,Y_MAX-1
	jl		bounce_success
	and		bl,not PROJ_DOWN
	or		bl,PROJ_UP
bounce_success:
	or		[sflags],SBOUNCE
bouncee:
	cmp		dl,X_LEFT
	jg		bouncew
	and		bl,not PROJ_LEFT
	or		bl,PROJ_RIGHT
	jmp		bounce_success2
bouncew:
	cmp		dl,X_RIGHT-1
	jl		bounce_end
	and		bl,not PROJ_RIGHT
	or		bl,PROJ_LEFT

bounce_success2:
	or		[sflags],SBOUNCE

bounce_end:
	ret

createprojectile:
; A projectile of a certain type is created.
; Call with:
;	AH = color
;	AL = character
;	BH = ticks to move (upper half of byte)
;	BL = direction
;	CX = count of objects at buffer address
;	DX = X and Y coordinates
;	SI = address of projectile data, subtracted by OBJ_LEN
; Reusable: CX, SI

	; Find the first suitable memory location available
	add		si,OBJ_LEN
	test		byte [si+1],0FFh
	loopnz		createprojectile
	jz		lodprojdata
	jmp		stopcreate
	
	; Load it up...
lodprojdata:
	mov		[si],ax
	mov		[si+2],dx
	mov		[si+4],bx
stopcreate:
	ret

checkticks:
; Unpack a byte and determine if the number of ticks has elapsed.
; Call with:
;	AL = packed tick counter
; Destroys: AX
; Preserves: CX
; Returns: ZF (if lower half tick counter matches upper half)
	mov		ah,al
	and		al,0Fh
	shr		ah,1
	shr		ah,1
	shr		ah,1
	shr		ah,1
	cmp		al,ah
	ret
	
enemyaction:
; Enemies are moved, created, or commanded to fire.
	mov		si,eattr
	
enemyexists:
	; Check if enemy exists here
	mov		dx,word [si+2]				; X and Y positions
	test		dx,dx
	jnz		enemyshotbyplayer
	jmp		nextenemy
	
enemyshotbyplayer:
	; Player's bullet hit enemy?
	mov		bx,si					; save SI in register (this is faster)
	mov		cx,PBULLET_CT
	mov		si,pbullet+2-OBJ_LEN
	call		checkcollide
	jz		r1
	; Player's bomb hit enemy?
	test		[gflags],4				; Don't try to check if one is not going off
	jnz		r0
	jmp		emove_restoreregs
r0:
	mov		cx,PBOMB_DATA_LEN
	mov		si,pbomb+2-OBJ_LEN
	call		checkcollide
	jz		r1
	jmp		emove_restoreregs
r1:
	cmp		si,pbomb				; Don't destroy bomb smoke
	jae		r2
	mov		word [si-2],0				; Hit! Destroy the bullet...
	mov		word [si],0
r2:
	mov		ax,si					; Save the new SI, need to know how much
								; damage to apply
	mov		si,bx
	dec		byte [si+6]				; Enemy took a hit...
	jz		enemyactkill
	cmp		ax,pbomb
	jb		enemynotkill
	sub		byte [si+6],7				; Bombs hit much harder
	jz		enemyactkill
	jc		enemyactkill
enemynotkill:
	; Turn it red briefly to indicate a hit was taken
	call		setpos
	mov		al,byte [si]
	HIT_COLOR
	call		safe_putch
	jmp		moveenemy
enemyactkill:
	mov		bh,[si+7]
	mov		ch,bh
	call		killenemy
	mov		bh,ch
	test		bh,80h					; Was this spawned after timeout?
	jnz		enemydonekilling			; If so, don't count towards quota
	dec		[eremain]
	; If zero, the wave is finished.
	jnz		enemyremain
	or		[gflags],2
	jmp		stopenemyaction

enemyremain:
	; Is there another enemy in the queue?
	cmp		[equeue],0
	jz		enemydonekilling			; No

	push		si
	dec		[equeue]
	mov		si,[eindex]
	mov		bh,[si]
	mov		dx,[si+1]

; TODO: Check if it overlaps another enemy.
; (that was a concern back then, but actually I like the effect of
; "cell division" it produces)
	call		safetycheck				; Don't spawn too close to the player!

	add		si,WAVE_EDATA_LEN
	mov		[eindex],si
	call		createenemy
	pop		si
	jmp		nextenemy

enemydonekilling:
	; If we're at Wave 25 and the timeout has not set in yet,
	; keep rolling them in.
	cmp		[wave],LAST_WAVE
	jb		enemydonekilling2
	test		[gflags],8
	jnz		enemydonekilling2
	mov		[equeue],255
	mov		[eindex],wave25+2
enemydonekilling2:
	jmp		nextenemy

emove_restoreregs:
	mov		si,bx
	
moveenemy:
	mov		bl,byte [si+4]				; Direction

	; Dashing enemies (and other four-ways) should respond instantaneously
	mov		al,byte [si+7]
	and		al,not 80h
	cmp		al,ENEMY_DASH
	jb		moveenemy2
	cmp		al,ENEMY_BLINKDASH
	ja		moveenemy2
me_dashing:
	test		bl,10h					; Already dashing?
	jnz		moveenemy2
	mov		cl,byte [si+8]
	call		dash_movement				; No, dash now
	test		bl,10h
	jnz		moveenemy3
	jmp		moveenemy2
moveenemy2:
	; Check if ready to move/act
	mov		al,byte [si+5]
	call		checkticks
	jne		enemytick				; Not ready
	and		byte [si+5],0F0h			; Ready, reset current ticks
	
moveenemy3:	
	call		setpos					; Clear existing character
	xor		ax,ax					; at original position
	call		safe_putch

	mov		ax,dx
	call		moveobject
	call		collidewall
	jz		enemysetoldpos
	call		wormcollide
	jz		enemysetoldpos
moveenemy4:
	push		si
	mov		cx,ENEMY_CT
	mov		si,eattr+2-OBJ_LEN
	call 		checkcollide
	pop		si
	jnz		enemysetnewpos
enemysetoldpos:
	mov		dx,ax
	mov		ch,1					; Indicate no move was made
	jmp		enemyputch
enemysetnewpos:
	mov		word [si+2],dx
enemyputch:
	call		adjpos
	mov		ax,word [si]				; Color and character data
	call		safe_putch
	
hitplayer:
	cmp		dx,word [player_x]
	jne		enemycd
	; Player dies routine goes here (mind character writes!)
	; (set a variable instead?)
	or		[gflags],1
	
	; Change direction?
enemycd:
	call		enemychdir
	jmp		nextenemy
	
enemytick:
	inc		byte [si+5]
	
nextenemy:
	xor		ch,ch
	add		si,OBJ_LEN
	cmp		si,eattr_end
	jae		stopenemyaction
	jmp		enemyexists
stopenemyaction:
	ret

enemychdir:
; Change the direction of the enemy.
; Call with:
;	SI = address of current enemy
;	BL = Auxiliary Halfbyte/Direction
;	DX = enemy location
; Returns: BL (stored to memory)
; Destroys: AX, BX, CX, DX (varies)
	mov		al,byte [si+7]				; Enemy type
	and		al,not 80h
	cmp		al,ENEMY_DASH
	jb		cdnormal
	cmp		al,ENEMY_BLINKDASH
	ja		cdnormal

cddash:
	test		bl,10h					; already dashing?
	jz		cddashstart

	; Check if the object is about to hit a wall
	call		collidewall_preempt
	jz		cddashend
	mov		cl,byte [si+8]
	call		dash_movement
	test		bl,10h
	jnz		cddashend

cddashstart:
	; Restore old tick rate from original enemy data
	; (we will get a new value for BL)
	;mov		bx,estupid+(ENEMY_DATA_LEN*(ENEMY_DASH-1))+2
	mov		bx,estupid-ENEMY_DATA_LEN
cddash_nextenemy:
	add		bx,ENEMY_DATA_LEN
	cmp		bx,enemylist_end
	jae		cddashend
	cmp		byte [bx],al
	jne		cddash_nextenemy
	
	mov		al,byte [bx+1]
	mov		byte [si+5],al
	xor		bl,bl
	mov		cx,si
	mov		si,rand4
	call		stupid_movement
	mov		si,cx
cddashend:
	jmp		enemycdend

cdnormal:
	mov		bh,bl
	and		bl,0F0h
cdstupid:
	; TODO: Organize the equates better so that we can use jae and jb
	cmp		al,ENEMY_STUPID
	jb		cdchase
	cmp		al,ENEMY_HYPERSTUPID
	ja		cdchase
	mov		cx,si					; Save old SI register value
	mov		ax,word [si+8]
	mov		si,ax
	call		stupid_movement
	mov		si,cx					; Restore old SI value
	jmp		enemycdend

cdchase:
	cmp		al,ENEMY_CHASE
	jb		cdgunner
	cmp		al,ENEMY_EVILCHASE
	ja		cdgunner
cdchase_execute:
	call		chase_movement
	jmp		enemycdend

; The reason why we need two extra bytes in the enemy objects.
cdgunner:
	cmp		al,ENEMY_GUNNER
	jb		cdworm
	cmp		al,ENEMY_DEATHLASER
	ja		cdworm
	mov		cx,si
	mov		si,rand3
	xor		ah,ah
	;add		si,ax					; adding __SLIGHT__ randomness...
	; First, move about somewhere
	call		stupid_movement
	mov		si,cx
	mov		bh,byte [si+8]
	mov		ah,byte [si+9]
	; Now... FIRE!
	add		bl,10h
	cmp		bl,ah
	jae		cdgunner2				; Don't fire if not ready...
	jmp		enemycdend
cdgunner2:
	call		enemy_shoot
	and		bl,0Fh					; Reset alternate tick counter
	jmp		enemycdend

cdworm:
	cmp		al,ENEMY_WORM
	jb		enemycdend
	cmp		al,ENEMY_MEGAWORM
	ja		enemycdend

	;test		bl,0Fh					; Avoid this completely if no direction is set
	;jz		enemycdend

	xor		bl,bl
	;call		chase_movement
	call		chase_fourway
	;cmp		ch,1
	;je		enemycdend
	xor		ch,ch
	mov		cl,byte [si+8]
	call		worm_movement
	;jmp		enemycdend

enemycdend:
	mov		byte [si+4],bl
	ret

collidewall_preempt:
; Similar to collidewall, except this preemptively changes the direction
; before the object is about to hit the wall.
; Call with:
;	DX = location
; Returns: BL, ZF (if about to hit the wall)
; Destroys: CH
	xor		ch,ch

	test		bl,PROJ_LEFT
	jz		cwp_right
	cmp		dl,X_LEFT+2
	jle		cwp_fail
	jmp		cwp_end
cwp_right:
	test		bl,PROJ_RIGHT
	jz		cwp_up
	cmp		dl,X_RIGHT-2
	jge		cwp_fail
	jmp		cwp_end
cwp_up:
	test		bl,PROJ_UP
	jz		cwp_down
	cmp		dh,2
	jle		cwp_fail
	jmp		cwp_end
cwp_down:
	test		bl,PROJ_DOWN
	jz		cwp_end
	cmp		dh,Y_MAX-2
	jle		cwp_end

cwp_fail:
	inc		ch

cwp_end:
	test		ch,ch
	ret

createenemy:
; Create a new enemy if space is available.
; Call with:
;	BH = enemy type
;	DX = X and Y positions
; Destroys: AX, DX, SI
; Preserves: CX
	mov		si,eattr
	push		cx
createenemy2:
	; check that this location is not already occupied...
	cmp		byte [si],0
	jnz		nextenemydata

	push		bx
	push		si
	mov		si,estupid
compareenemy:
	mov		al,byte [si]
	cmp		bh,al				; Check if this is the enemy type we need
	je		stopcompareenemy
	add		si,ENEMY_DATA_LEN
	jmp		compareenemy
stopcompareenemy:
	mov		ax,word [si+3]			; ASCII data
	mov		cx,word [si+1]			; hit points, ticks to move
	mov		bx,word [si+7]			; auxiliary data
	pop		si

	mov		word [si+8],bx
	pop		bx
	mov		byte [si+7],bh			; enemy type
	mov		word [si],ax
	mov		word [si+2],dx			; initial coordinates
	mov		word [si+5],cx

	pushf						; Tell other processes we found an open space afterwards!
	; Wave 25 enemies are HARDER!!!
	cmp		[wave],LAST_WAVE
	jne		createenemy_disp
	add		byte [si+6],3			; More hits to take
	sub		byte [si+5],20h			; Less ticks to move
createenemy_disp:
	; Show On Screen Now
	call		setpos
	call		safe_putch
	popf
	jmp		stopcreateenemy

nextenemydata:
	add		si,OBJ_LEN
	cmp		si,eattr_end
	jb		createenemy2
	test		si,si				; clear zero flag
	
stopcreateenemy:
	pop		cx
	ret
	
killenemy:
; Destroys an enemy in memory and gives the player a points
; Call with:
;	BH = enemy type
; 	SI = enemy address
; Preserves: SI, CX
; Destroys: AX, DX
	and		bh,not 80h
	cmp		bh,ENEMY_WORM
	jb		killenemy2
	cmp		bh,ENEMY_MEGAWORM
	ja		killenemy2
	call		worm_kill

killenemy2:
	mov		word [si],0
	mov		word [si+2],0
	or		[sflags],SEDEATH
	mov		[stimer],0
	
	; Ensure there's no residue when the enemy dies
	call		setpos
	xor		ax,ax
	call		safe_putch	

	; Award Points
	push		si
	push		cx
	mov		si,estupid-ENEMY_DATA_LEN
	mov		cx,NUMBER_OF_ENEMIES
nextenemytokill:
	add		si,ENEMY_DATA_LEN
	cmp		bh,byte [si]
	loopne		nextenemytokill
	; Load the enemy's score and add to the player's
	mov		ax,word [si+5]
	cmp		[wave],LAST_WAVE
	jb		normal_score
	shl		ax,1				; Wave 25 offers double scoring
normal_score:
	call		addpoints
	call		updatehud

	pop		cx
	pop		si

	ret

addpoints:
; Add points to the player's score.
; Call with:
;	AX = score to add
; Destroys: AX
	; Add the points (has to be addressed 16 bits at a time
	; to avoid crashes on 8088/286)
	add		word [score],ax
	adc		word [score+2],0
	jc		stopwrap

	; Do not offer anymore bonuses after 1,000,000 points.
	cmp		word [score+2],10h
	jae		addpoints_end
	cmp		word [score+2],0Fh
	jb		addbonus
	cmp		word [score],4240h
	jae		addpoints_end
	
addbonus:
	add		word [auxscore],ax

	; Is the player eligible to receive Free Benefits
	cmp		word [auxscore],BONUS_MAN
	jb		addpoints_end

	sub		word [auxscore],BONUS_MAN
	inc		[lives]
	inc		[bombs]
	push		ax
	; There was a bonus. Let's cram in a cool sound effect!
	mov		ax,PIT_FREQ/1000
	mov		cx,7
	mov		dx,400
	call		sweep_sound
	pop		ax
	cmp		al,0FFh						; set carry flag used by updatehud
	jmp		addpoints_end

stopwrap:
	; If the player somehow manages to max out the 32-bit integer, well...
	; CORNGRATULATION YOU HAVE NO LIFE
	mov		word [score],0FFFFh
	mov		word [score+2],0FFFFh

addpoints_end:
	ret

updatehud:
; Update the HUD at the top of the screen with the current score.
; Call with:
;	CF = determines whether to add a bomb and life to the HUD
;		(assumed to be called directly after addpoints)
; Destroys: AX, BX, CX, DX
	jnc		updatescore

	; Add another life icon (if there is room)
	xor		ah,ah
	mov		al,[lives]
	dec		al
	test		al,al
	jz		addbomb
	cmp		al,7
	ja		addbomb
	shl		al,1
	mov		di,LIVES_POS-2+(X_LEFT*2)
	add		di,ax
	mov		al,PLAYER_CH
	stosb

addbomb:
	; Add another bomb icon (if there is room)
	xor		ah,ah
	mov		al,[bombs]
	test		al,al
	jz		updatescore
	cmp		al,8
	ja		updatescore
	shl		al,1
	mov		di,BOMBS_POS-2+(X_LEFT*2)
	add		di,ax
	mov		al,ASC_BOMB
	stosb

updatescore:
	mov		si,score
	mov		di,SCORE_POS-(X_LEFT*2)
	call		printnum32
	
	ret

timeout_spawner:
TIMEOUT_ITERATION = 48
; Spawn new enemies as the timer runs out.
	cmp		[timeout],TIMEOUT_ITERATION				; spawn every number of ticks in TIMEOUT_ITERATION
	jb		timeout_end
	sub		[timeout],TIMEOUT_ITERATION
	mov		si,[etindex]

	; Check if we hit the end of the wave enemy data
	cmp		[etcount],0
	jnz		timeout_dospawn

	; Start using Wave 25 data to really throw them off
	mov		si,wave25+2
	mov		[etindex],si
	mov		[etcount],255

timeout_dospawn:
	mov		bh,byte [si]
	mov		dx,word [si+1]
	call		safetycheck
	call		createenemy
	jnz		timeout_end
	; The new enemy must be slightly harder, and always colored bright red.
	; A macro will determine if bright red or bright white should be used
	; depending on if the MDA version is being built or not.
	TIMEOUT_COLOR
	add		byte [si+6],2					; More hits to take
	sub		byte [si+5],20h					; Less ticks to move
	or		byte [si+7],80h					; Not part of the blood quota, just there to annoy you
	add		[etindex],WAVE_EDATA_LEN
	dec		[etcount]

timeout_end:
	inc		[timeout]
	ret

safetycheck:
; SAFELY spawn the enemy far enough from the player to avoid
; cheap collisions.
; Call with:
;	DX = location of new enemy
; Destroys: AX
; Returns: DX
SAFETY_RADIUS	= 9
	mov		ax,word [player_x]
	sub		ax,(SAFETY_RADIUS*256)+SAFETY_RADIUS

	cmp		dl,al
	jle		saferight
	mov		dl,al
	cmp		dl,X_LEFT+1
	jg		safeup
	add		dl,SAFETY_RADIUS*2
	jmp		safeup
saferight:
	add		al,SAFETY_RADIUS*2
	cmp		dl,al
	jg		safeup
	mov		dl,al
	cmp		dl,X_RIGHT-1
	jl		safeup
	sub		dl,SAFETY_RADIUS*2

safeup:
	;sub		ah,SAFETY_RADIUS
	cmp		dh,ah
	jle		safedown
	mov		dh,ah
	cmp		dh,1
	jg		safeend
	add		dh,SAFETY_RADIUS*2
	jmp		safeend
safedown:
	add		ah,SAFETY_RADIUS*2
	cmp		dh,ah
	jg		safeend
	mov		dh,ah
	cmp		dh,Y_MAX-1
	jl		safeend
	sub		dh,SAFETY_RADIUS*2

safeend:
	ret

printnum32:
; Display a number on the screen (aligned right to left).
; Expects a number that uses BCD format.
; Call with:
;	SI = number to display
;	DI = location to print number
; Destroys: AX, BX, CX, DX
	
	mov		byte [tmpbuf],0
	mov		dx,[si+2]
	mov		ax,[si]
	
	; Routine to split the number into decimal digits,
	; I didn't come up with this one... merely adapted it.
	mov		bx,10
	push		bx
pn1:
	mov		cx,ax
	mov		ax,dx
	xor		dx,dx
	div		bx
	xchg		ax,cx
	div		bx
	push		dx
	mov		dx,cx
	inc		byte [tmpbuf]
	or		cx,ax
	jnz		pn1
	pop		dx
	mov		ah,byte [es:di+1]
	
	; Restore digit count
	mov		al,byte [tmpbuf]
	
	; Determine the destination index
	mov		cx,10
	sub		cl,al
	shl		cl,1
	add		di,cx
pn2:
	add		dl,'0'
	mov		al,dl
	call		safe_putch
	pop		dx
	cmp		dx,bx
	jb		pn2
	ret

playerbomb:
; Assumes the player hit SPACE to detonate a bomb, hitting many
; enemies within the specified diameter. This will continue to be called
; until the smoke clears out.
	
	mov		bh,20h					; set the speed of the bomb explosion
	mov		si,pbomb
	cmp		[pbomb_timer],0
	jz		new_playerbomb

	; The bomb already exists, keep expanding its explosion up to a point
	mov		ah,15
	call		bombexpand
	jnz		playerbomb_end
	and		[gflags],not 4
	; A procedure should remove the existing particles.
	jmp		playerbomb_end

new_playerbomb:
	mov		al,[bombs]
	test		al,al
	jz		playerbomb_fail				; don't do anything if bomb stock is depleted
	dec		al
	mov		[bombs],al

	cmp		al,8
	jae		playerbomb_go				; do not remove an icon if there is still 8 or more
	; Remove a bomb character from the HUD
	mov		di,BOMBS_POS+(X_LEFT*2)
	xor		ah,ah
	shl		al,1
	add		di,ax
	xor		al,al
	stosb

playerbomb_go:
	mov		[pbomb_timer],PBOMB_RADIUS
	mov		ah,15
	mov		dx,word [player_x]
	sub		si,OBJ_LEN
	call		bombset
	or		[sflags],SBOMB
	mov		[stimer],0
	jmp		playerbomb_end

playerbomb_fail:
	and		[gflags],not 4
playerbomb_end:
	ret

bombexpand:
; Expand the bomb's explosion.
; Call with:
;	AH = player bomb color
;	BH = ticks to move (upper half of byte)
;	SI = address of bomb data
; Destroys: AX, BX, CX, DX
; Returns: ZF (if radius exceeded)

	; Is there an explosion going on?
	cmp		[pbomb_timer],0
	jnz		bexpand_ticks
	mov		al,1
	jmp		bexpand_end

bexpand_ticks:
	; Check if these are ready to move...
	mov		al,[pbomb_timer2]
	push		ax
	call		checkticks
	pop		ax
	jz		bexpand_start
	inc		[pbomb_timer2]
	jmp		bexpand_end

bexpand_start:
	and		[pbomb_timer2],0F0h
	; Don't make it grow too far...
	; NOTE: maybe not make that rely on an equate
	mov		bl,[pbomb_timer]
	dec		bl
	jz		destroybomb
bexpand_ok:
	mov		[pbomb_timer],bl
	; Upwards explosion	
	mov		al,ASC_FULLBLOCK

	; There should be a nice effect where the "smoke" dissipates.
	mov		cx,PBOMB_DATA_LEN

	cmp		bl,3					; Light smoke
	ja		medium_explode
	mov		al,ASC_LIGHTBLOCK
	jmp		bomb_rpch
medium_explode:
	cmp		bl,5					; Medium smoke
	ja		dark_explode
	mov		al,ASC_MEDIUMBLOCK
	jmp		bomb_rpch
dark_explode:
	cmp		bl,7					; Dark smoke
	ja		bomb_newsmoke
	mov		al,ASC_DARKBLOCK
bomb_rpch:
	call		replicate_char

bomb_newsmoke:
	add		si,8*OBJ_LEN				; Only work with space after the initial objects
	sub		cx,9
	; All new objects are plotted in a clockwise fashion.
	; Bombs use the "ripple" procedure in four directions.
	mov		dx,word [si+2-(8*OBJ_LEN)]
	test		dx,dx
	jz		bomb_smokeright
	mov		bl,PROJ_UP
	call		ripple

bomb_smokeright:
	mov		dx,word [si+2-(6*OBJ_LEN)]
	test		dx,dx
	jz		bomb_smokedown
	mov		bl,PROJ_RIGHT
	call		ripple

bomb_smokedown:
	mov		dx,word [si+2-(4*OBJ_LEN)]
	test		dx,dx
	jz		bomb_smokeleft
	mov		bl,PROJ_DOWN
	call		ripple

bomb_smokeleft:
	mov		dx,word [si+2-(2*OBJ_LEN)]
	test		dx,dx
	jz		bexpand_end
	mov		bl,PROJ_LEFT
	call		ripple
	jmp		bexpand_end

destroybomb:
	mov		[pbomb_timer],0
	xor		ax,ax
	mov		cx,PBOMB_DATA_LEN
	push		si
destroybomb2:
	; Clear out existing characters on the screen
	mov		dx,word [si+2]
	test		dx,dx					; Don't overwrite the upper right corner
	jz		destroybomb3
	call		setpos
	call		safe_putch
destroybomb3:
	add		si,OBJ_LEN
	loop		destroybomb2
	pop		si
	mov		cx,(PBOMB_DATA_LEN)*(OBJ_LEN/2)
	call		initbuf

bexpand_end:
	test		ax,ax					; Set or clear zero flag
	ret

ripple:
; Create a "rippling" projectile by duplicating it on two sides next to it.
; The objects move diagonally away from the source.
; The ripple effect is achieved by calling this repeatedly.
; Call with:
;	AX = character and color information
;	BH = projectile speed
;	BL = direction
;	CX = count of objects available in data address
;	DX = location of existing object to ripple
;	SI = object address
; Preserves: BX, CX, SI
	push		si
	push		bx
	push		cx
	and		bl,0Fh
ripple_n:
	cmp		bl,PROJ_UP
	jne		ripple_ne
	mov		bl,PROJ_UP+PROJ_LEFT
	push		bx				; Save this direction as the second projectile
	mov		bl,PROJ_UP+PROJ_RIGHT
	jmp		ripple_create
	
ripple_ne:
	cmp		bl,PROJ_UP+PROJ_RIGHT
	jne		ripple_e
	mov		bl,PROJ_RIGHT
	push		bx
	mov		bl,PROJ_UP
	jmp		ripple_create

ripple_e:
	cmp		bl,PROJ_RIGHT
	jne		ripple_se
	mov		bl,PROJ_UP+PROJ_RIGHT
	push		bx
	mov		bl,PROJ_DOWN+PROJ_RIGHT
	jmp		ripple_create

ripple_se:
	cmp		bl,PROJ_DOWN+PROJ_RIGHT
	jne		ripple_s
	mov		bl,PROJ_RIGHT
	push		bx
	mov		bl,PROJ_DOWN
	jmp		ripple_create

ripple_s:
	cmp		bl,PROJ_DOWN
	jne		ripple_sw
	mov		bl,PROJ_DOWN+PROJ_LEFT
	push		bx
	mov		bl,PROJ_DOWN+PROJ_RIGHT
	jmp		ripple_create

ripple_sw:
	cmp		bl,PROJ_DOWN+PROJ_LEFT
	jne		ripple_w
	mov		bl,PROJ_DOWN
	push		bx
	mov		bl,PROJ_LEFT
	jmp		ripple_create

ripple_w:
	cmp		bl,PROJ_LEFT
	jne		ripple_nw
	mov		bl,PROJ_DOWN+PROJ_LEFT
	push		bx
	mov		bl,PROJ_UP+PROJ_LEFT
	jmp		ripple_create

; Could also just assume northwest by default but it might look stupid if
; something goes wrong.
ripple_nw:
	cmp		bl,PROJ_UP+PROJ_LEFT
	jne		ripple_end
	mov		bl,PROJ_LEFT
	push		bx
	mov		bl,PROJ_UP

ripple_create:
	call		createprojectile
	pop		bx
	test		cx,cx					; If there is no space left, do not make another one
	jz		ripple_end
	call		createprojectile

ripple_end:
	pop		cx
	pop		bx
	pop		si
	ret

bombset:
; A new bomb is planted. Cool!
; Call with:
;	AH = color of bomb smoke
;	BH = ticks to move (upper half of byte)
;	DX = origin of of Bom
;	SI = address of bomb data
; Destroys: AX, BX, CX

	mov		al,ASC_FULLBLOCK

	mov		cx,8					; fire in all eight directions
								; (Other procedures must expand the death smoke)
bomb_putparticle:
	push		cx
	cmp		cl,8
	jb		bsetne
	mov		bl,PROJ_UP
	jmp		bombcall
bsetne:
	cmp		cl,7
	jb		bsete
	mov		bl,PROJ_UP+PROJ_RIGHT
	jmp		bombcall
bsete:
	cmp		cl,6
	jb		bsetse
	mov		bl,PROJ_RIGHT
	jmp		bombcall
bsetse:
	cmp		cl,5
	jb		bsets
	mov		bl,PROJ_DOWN+PROJ_RIGHT
	jmp		bombcall
bsets:
	cmp		cl,4
	jb		bsetsw
	mov		bl,PROJ_DOWN
	jmp		bombcall
bsetsw:
	cmp		cl,3
	jb		bsetw
	mov		bl,PROJ_DOWN+PROJ_LEFT
	jmp		bombcall
bsetw:
	cmp		cl,2
	jb		bsetnw
	mov		bl,PROJ_LEFT
	jmp		bombcall
bsetnw:
	mov		bl,PROJ_UP+PROJ_LEFT

bombcall:
	mov		cx,PBOMB_DATA_LEN
	push		si
	call		createprojectile
	pop		si
	pop		cx
	loop		bomb_putparticle
	ret

clear_screen:
; Clear the screen
	xor		ax,ax
	xor		di,di
	mov		cx,X_MAX*Y_MAX
	rep		stosw
	ret

replicate_char:
; Replicate one character across an entire data address.
; Call with:
;	AL = character to replicate
;	CX = count of objects at address
;	SI = address of object data
; Preserves: CX, SI
	push		cx
	push		si
rccmpch:
	cmp		byte [si],0			; Don't do this on nonexistent objects
	jz		rcnextch
	mov		byte [si],al
	add		si,OBJ_LEN
rcnextch:
	loop		rccmpch
	pop		si
	pop		cx
	ret

checkwavetime:
WARNING_TIME = 600
; Compares gtimer against wave_time to see if the wave's time limit
; was exceeded.
; Call with:
;	DX = value of gtimer (should have been popped around the end of main loop)
; Destroys: AX, BX, CX
	; Don't even bother if we're on the last wave, everything
	; is hyperintensified as is.
	cmp		[wave],LAST_WAVE
	je		not_expired
	mov		bl,[gflags]
	test		bl,8				; Don't overwrite the HUD over and over again
	jnz		not_expired
comparetime:
	mov		cx,[wave_time]
	cmp		dx,cx
	jb		timewarn
	REDHUD
	jmp		settimeout

timewarn:
	sub		cx,WARNING_TIME
	cmp		dx,cx
	jb		not_expired
	mov		al,6Fh				; Turn it yellow/orange in warning mode
	jmp		hudcolor

settimeout:
	or		bl,8
	; Turn the HUD bar red
hudcolor:
	mov		di,1+(X_LEFT*2)
	cmp		al,byte [es:di]			; Again, don't repeat this so much
	je		not_expired
	mov		cx,WIDTH
hudcolor2:
	stosb
	inc		di
	loop		hudcolor2
	or		[sflags],STIME
	test		bl,8
	jz		not_expired
	mov		[gflags],bl

not_expired:
	ret

printstr:
; Print a string directly to video memory.
; Call with:
;	AH = text color
;	SI = pointer to string
;	DI = position of string
; Destroys: CX
	push		di

printstr_check:
	lodsb
	cmp		al,10				; Line break?
	jne		printstr_putch
	pop		di
	add		di,X_MAX*2			; Advance to next line, same first column
	push		di
	jmp		printstr_check

printstr_putch:
	call		safe_putch
	test		byte [si],0FFh
	jnz		printstr_check
	pop		di
	ret
	
drawbox:
; Draw a double-bordered ASCII box.
; Call with:
;	AH = border color
;	CH = horizontal size
;	CL = vertical size
;	DH = vertical position
;	DL = horizontal position
; Destroys: BX

	; Do not count bordering lines for some of the upcoming loops
	sub		cx,0202h
	
	; Assign a position for the box
	call		setpos
	mov		bx,di
	
	; Draw upper left corner of border
	mov		al,0C9h
	call		safe_putch
	; Draw top horizontal border
	mov		al,0CDh
	push		cx
	xchg		cl,ch
	xor		ch,ch
	rep		stosw
	; Draw upper right corner of border
	mov		al,0BBh
	call		safe_putch
	
	; Move DI to exact position underneath top of border
	pop		cx
	mov		di,bx
	add		di,X_MAX*2
	
	; Draw vertical borders on each side
	call		dvertbar
	xor		dh,dh
	mov		dl,ch
	shl		dl,1
	add		bx,dx
	add		bx,(X_MAX+1)*2
	mov		di,bx
	call		dvertbar
	mov		dl,ch
	inc		dl
	shl		dl,1
	sub		di,dx
	
	; Draw bottom left corner of border
	mov		al,0C8h
	call		safe_putch
	; Draw bottom horizontal border
	mov		al,0CDh
	push		cx
	xchg		cl,ch
	xor		ch,ch
	rep		stosw
	pop		cx
	; Draw bottom right corner of border
	mov		al,0BCh
	call		safe_putch
	
	ret

dvertbar:
	; Draw a vertical bar
	push		cx
	xor		ch,ch
	mov		al,0BAh
vbloop:
	call		safe_putch
	add		di,(X_MAX-1)*2
	loop		vbloop
	pop		cx
	ret

wavemsg:
; Display a brief message pertaining to the wave...
; Called when a new wave starts.

	; The extra segment is briefly changed to exploit the
	; convenience of rep, rather than using SI and loop.
	push		es
	mov		ax,ds
	mov		es,ax
	mov		di,wavemsgstart			; Contains only a 0, used so this
							; routine can land on the required message accurately

	; First we need to find the right message. This can be done
	; by effectively counting through null byte terminators.
	xor		al,al				; We need to find a null terminator
	xor		ch,ch
	mov		cl,[wave]
findwavemsg:
	push		cx
	mov		cx,200				; This looks like a safe number.
	repne		scasb
	pop		cx				; Restore the wave number
	loop		findwavemsg
	mov		si,di				; And we have our string
	pop		es

	; Now, clear out an area for the message box...
	mov		bl,6
	mov		cl,38
	mov		dx,(8*256)+1+X_LEFT
	call		box_clear

	; ...then print the string.
	mov		dx,(10*256)+3+X_LEFT
	call		setpos
	mov		ah,15
	call		printstr

FLICKER_RATE = 6
	mov		cx,100
	mov		ax,PIT_FREQ/100
	mov		bx,PIT_FREQ/120
	mov		dx,50
	mov		[tmpbuf],(7*256)+FLICKER_RATE
wavemsgsound:
	push		cx
	mov		cx,2
	push		bx
	call		sweep_sound
	pop		bx
	pop		cx
	sub		ax,300
	xchg		ax,bx

	; Make the border trippy!
	; (P.S. you can't use pusha on 8086)
	push		ax
	push		bx
	push		cx
	push		dx
	mov		ax,[tmpbuf]
	inc		al
	cmp		al,FLICKER_RATE
	jb		wavemsg_finishborder
	mov		cx,(38*256)+6
	mov		dx,(8*256)+1+X_LEFT
	call		drawbox
	xor		al,al
	WAVEMSG_FLICKER
wavemsg_finishborder:
	mov		[tmpbuf],ax
	pop		dx
	pop		cx
	pop		bx
	pop		ax

	loop		wavemsgsound
wavemsg_end:
	ret

; -- LEVEL INITIALIZATION ROUTINE --
; Called during normal gameplay.
init_level:
	; Kill all existing sound
	call		sound_kill

	call		clear_screen
	mov		[gflags],0

	mov		[player_x],START_X
	mov		[player_y],START_Y

	WAVE_BORDER
	mov		cx,Y_MAX+(WIDTH*256)-1		; One row shorter than the screen
	mov		dx,100h+X_LEFT
	call		drawbox
	
	; display HUD at top
	mov		di,X_LEFT*2
	HUDCOLOR
	GREENHUD

hudbar:
	mov		cx,WIDTH
	rep		stosw

	; Lives counter
	mov		di,LIVES_POS+(X_LEFT*2)
	mov		cl,[lives]
	dec		cl
	cmp		cl,8
	jb		putlives
	mov		cl,7
putlives:
	mov		al,PLAYER_CH
	rep		stosw

	; Bomb counter
	mov		di,BOMBS_POS+(X_LEFT*2)
	mov		cl,[bombs]
	cmp		cl,9
	jb		putbombs
	mov		cl,8
putbombs:
	mov		al,ASC_BOMB
	rep		stosw
	
	; Wave counter
	mov		di,(16+X_LEFT)*2
	mov		si,wavestr
	call		printstr
	mov		di,(16*2)+(5*2)+(X_LEFT*2)			; explicitly set new position for wave count
									; (last-minute change to printstr requires this)
	
	xor		ah,ah
	mov		al,[wave]
	mov		bl,10
	div		bl
	test		al,al
	jz		nextwavedigit
	push		ax
	add		al,'0'
	stosb
	inc		di
	pop		ax
nextwavedigit:
	mov		al,ah
	add		al,'0'
	stosb
	
	; Display score
	mov		ah,1Fh
	mov		si,score
	mov		di,SCORE_POS-(X_LEFT*2)
	call		printnum32
	
	; Initialize some hefty temporary buffers
	mov		cx,PBULLET_CT*(OBJ_LEN/2)	; Four quadwords
	mov		si,pbullet					; Player bullets
	call		initbuf
	
	mov		cx,ENEMY_CT*(OBJ_LEN/2)
	mov		si,eattr					; Enemy attributes
	call		initbuf

	mov		cx,PROJ_CT*(OBJ_LEN/2)
	mov		si,eproj					; Enemy projectiles
	call		initbuf

	mov		cx,WORM_CT*(WORM_DATA_LEN/2)			; Worm body parts
	mov		si,eworm
	call		initbuf

	mov		[pbomb_timer],0
	mov		[pbomb_timer2],30h
	mov		cx,PBOMB_DATA_LEN*(OBJ_LEN/2)
	mov		si,pbomb					; Player bomb smoke
	call		initbuf

	; Align pointer to the correct wave number...
	mov		si,wave1+2
	xor		ch,ch
checkwavenumber:
	mov		cl,byte [si-1]					; Get number of enemies in wave
	push		cx
	xor		ah,ah						; Use this enemy count as the basis
	mov		al,cl						; for the wave timer
	mov		cl,5
	shl		ax,cl
	mov		[wave_time],ax
	pop		cx
	mov		al,byte [si-2]					; What is the wave number?
	cmp		al,[wave]
	je		startloadingwave
skipwavedata:
	add		si,WAVE_EDATA_LEN
	loop		skipwavedata
	add		si,2						; push ahead of the wave header
	jmp		checkwavenumber

startloadingwave:
	mov		[equeue],cl
	mov		[eremain],cl
	mov		[etcount],cl
	mov		[etindex],si
	mov		cl,ENEMY_CT
populate:
	; Populate memory with the enemies for this wave
	;push		cx
	mov		bh,byte [si]					; Load registers with data for the procedure
	mov		dx,word [si+1]
	call		safetycheck					; Avoid unsafe spawn locations
	push		si
	;mov		cx,ENEMY_CT
	call		createenemy
	pop		si
	;pop		cx
	dec		[equeue]
	add		si,3						; advance to next enemy index
	loop		populate

	mov		[eindex],si					; Save position of our enemy address for later
	
	; PSHAAAAW TRANSITION
	mov		si,eproj-OBJ_LEN
	mov		cx,4

TRANSITION_DISTANCE = 10
	mov		ax,(15*256)+ASC_TRIRIGHT
	mov		bx,PROJ_RIGHT
	mov		dx,(START_Y*256)+(START_X-TRANSITION_DISTANCE)
	call		createprojectile
	mov		al,ASC_TRILEFT
	mov		bl,PROJ_LEFT
	mov		dx,(START_Y*256)+(START_X+TRANSITION_DISTANCE)
	call		createprojectile
	mov		al,ASC_TRIDOWN
	mov		bl,PROJ_DOWN
	mov		dx,((START_Y-TRANSITION_DISTANCE)*256)+START_X
	call		createprojectile
	mov		al,ASC_TRIUP
	mov		bl,PROJ_UP
	mov		dx,((START_Y+TRANSITION_DISTANCE)*256)+START_X
	call		createprojectile
	mov		dx,[gtimer]

	; Create a sound also
	mov		ax,PIT_FREQ/40000
arrows_closein:
	; Little by little, move those arrows closer to the player's position
	; UNTIL they're immediately next to that.
	add		dx,8
	push		dx
	mov		si,eproj
	mov		cx,4
	push		ax
	call		moveprojectiles
	pop		ax
	mov		cx,7
	mov		dx,600
	call		sweep_sound
	pop		dx
arrows_closein2:
	cmp		[gtimer],dx
	jbe		arrows_closein2
	cmp		byte [eproj+2],START_X-1			; Last transitional character just under the player?
	jl		arrows_closein
	call		sound_kill
	
put_player:
	; Clear out the now useless transitional characters that "merged into" the player
	xor		ax,ax
	sub		di,2
	call		safe_putch
	sub		di,X_MAX*2
	call		safe_putch
	sub		di,6
	call		safe_putch
	sub		di,(X_MAX*2)
	call		safe_putch
	mov		cx,4*(OBJ_LEN/2)
	mov		si,eproj
	call		initbuf

	; Put the player at the
	mov		dx,word [player_x]
	call		setpos
	mov		ax,(15*256)+PLAYER_CH
	call		safe_putch
	mov		[timeout],0
	mov		[gtimer],0
	ret
	
; -- END OF LEVEL INITIALIZATION ROUTINE --

title:
; Display the title screen.
; This shows up when the user runs the program.
	TITLE_BORDER
	mov		cx,Y_MAX+(WIDTH*256)	; Take up the entire screen
	mov		dx,X_LEFT
	call		drawbox
	
	; Print the glorious title
	mov		ah,15
	mov		di,16+((X_MAX*2)*3)+(X_LEFT*2)
	mov		si,gname
	call		printstr

	; ...and its cheesy slogan	
	mov		di,10+((X_MAX*2)*5)+(X_LEFT*2)
	mov		si,tstr1
	call		printstr
	
	; Explain the controls and prompt user to start the bloodbath
	mov		di,4+((X_MAX*2)*10)+(X_LEFT*2)
	mov		si,tstr10
	call		printstr
	
	mov		di,8+((X_MAX*2)*11)+(X_LEFT*2)
	mov		si,tstr11
	call		printstr
	
	; Last few messages
	mov		di,8+((X_MAX*2)*16)+(X_LEFT*2)
	mov		si,tstr20
	call		printstr
	
	mov		di,2+((X_MAX*2)*20)+(X_LEFT*2)
	mov		si,tstr30
	call		printstr

	mov		di,(32*2)+((X_MAX*2)*23)+(X_LEFT*2)
	mov		si,version_str
	call		printstr
	ret

death_animation:
; Player death animation.

	; Play the sound first
	mov		ax,PIT_FREQ/900
	mov		cx,7
	mov		dx,600
	call		sweep_sound
	call		sound_kill

	mov		bx,[gtimer]
	add		bx,15
	mov		cx,4					; Number of loops
	mov		dx,word [player_x]
	call		setpos
	inc		di
death_loop:
	mov		al,7
	stosb
	dec		di
	mov		ax,PIT_FREQ/800
	mov		dx,700
	push		cx
	mov		cx,7
	call		sweep_sound
	call		gtimer_wait
	pop		cx
	add		bx,15

	mov		al,8
	stosb
	dec		di
	mov		ax,PIT_FREQ/700
	mov		dx,800
	push		cx
	mov		cx,7
	call		sweep_sound
	call		gtimer_wait
	pop		cx
	add		bx,15

	mov		al,15
	stosb
	dec		di
	mov		ax,PIT_FREQ/600
	mov		dx,900
	push		cx
	mov		cx,7
	call		sweep_sound
	call		gtimer_wait
	pop		cx
	add		bx,15

	loop		death_loop
	call		sound_kill
	ret

box_clear:
; Clear a box of a specific size.
; Don't use this if you want to clear the entire screen.
; Call with:
;	BL = box rows
;	CL = box columns
;	DX = coordinates of upper right corner of box
; Destroys: AX, BL, CX, DI
	call		setpos
	xor		ax,ax
	xor		ch,ch
box_clear_row:
	push		di
	push		cx
	rep		stosw
	pop		cx
	pop		di
	add		di,X_MAX*2
	dec		bl
	jnz		box_clear_row
	ret


gameover:
; The game has ended. An ample opportunity must be provided for the user
; to see their score.

	; First blank out the space for the box.
	mov		bl,5
	mov		cl,26
	mov		dx,(9*256)+7+X_LEFT
	call		box_clear

	; Display general strings first.
	mov		ah,15
	mov		dx,(10*256)+9+X_LEFT
	call		setpos
	mov		si,result_str
	call		printstr

	mov		dx,(11*256)+9+X_LEFT
	call		setpos
	mov		si,endmsg_str
	call		printstr

	mov		dx,(10*256)+9+13+X_LEFT
	call		setpos
	; "Defective"	= typical result not meeting any cool conditions
	; "Potential"	= scored above 500,000 but not wave 25
	; "Optimal"	= reached Wave 25
	; "Perfect"	= this and also scoring 1.5 million or more
	cmp		[wave],LAST_WAVE
	jb		gameover_potential

	cmp		word [score+2],17h
	jae		gameover_perfect
	cmp		word [score+2],16h
	jb		gameover_optimal
	cmp		word [score],0E360h
	jb		gameover_optimal
gameover_perfect:
	mov		ah,14				; Intense yellow
	mov		si,perfect_str
	call		printstr
	jmp		gameover_drawbox
gameover_optimal:
	mov		ah,10				; Intense green
	mov		si,optimal_str
	call		printstr
	jmp		gameover_drawbox
gameover_potential:
	cmp		word [score+2],8h
	jae		gameover_potential2
	cmp		word [score+2],7h
	jb		gameover_defective
	cmp		word [score],0A120h
	jb		gameover_defective
gameover_potential2:
	POTENTIAL_TEXT
	mov		si,potential_str
	call		printstr
	jmp		gameover_drawbox
gameover_defective:
	mov		ah,4				; Red
	mov		si,defective_str
	call		printstr

gameover_drawbox:
	mov		cx,(26*256)+5
	mov		dx,(9*256)+7+X_LEFT
	call		drawbox

gameover_waitrelease:
	; Wait for all keys to be lifted.
	test		[moveshootkeys],0FFh
	jnz		gameover_waitrelease
	ret

gtimer_wait:
; Call with:
;	BX = starting value of [gtimer]
; Destroys: BX
	cmp		bx,[gtimer]
	jae		gtimer_wait
	mov		bx,[gtimer]
	ret
	
initbuf:
; Initialize undefined buffers to all zeroes.
; Call with:
;	CX = number of words to initialize
;	SI = buffer
	mov		word [si],0
	add		si,2
	loop		initbuf
	ret

