include 'equ.asm'
include 'macros.asm'

; -- DISPLAY TITLE SCREEN --
program_start:
	call	clear_screen
	call	title
	
; Wait for input...
wait_title:
	call		sound_toggle_eval
	; Kill the program if escape pressed
	test		[otherkeys],1
	jz		wait_anykey
	jmp		exit

wait_anykey:	
	; Press any key to begin
	test		[moveshootkeys],0FFh
	jz		wait_title
	
	; assign default game settings
	mov		[gflags],0
	mov		word [score],0
	mov		word [score+2],0
	mov		word [auxscore],0
	mov		[lives],START_LIVES
	mov		[bombs],START_BOMBS
	mov		[wave],5
	
	call		init_level
	
; -- END OF DISPLAY TITLE SCREEN --

; **** Main loop ****
main:
	call		sound_toggle_eval

pausekey:
	; Pause requested?
	test		[otherkeys],4
	jz		bombkey
	call		pause_loop

bombkey:
	; Bomb key
	test		[otherkeys],2
	jz		main2
	or		[gflags],4
	and		[otherkeys],not 2

main2:
	mov		dx,[gtimer]
	push		dx

	; Player moves first
	call		moveplayer

	; Player shoots
	call		playershoot
	mov		si,pbullet
	mov		cx,PBULLET_CT
	call		moveprojectiles
	; Emergency bomb was set?
	test		[gflags],4
	jz		main_enemies
	call		playerbomb
	mov		si,pbomb
	mov		cx,PBOMB_DATA_LEN
	call		moveprojectiles

main_enemies:
	; Enemies move and do stuff
	call		enemyaction
	mov		si,eproj
	mov		cx,PROJ_CT
	call		moveprojectiles

	; Play sounds through PC speaker
	call		sound_loop

	test		[gflags],8			; don't call it if time already expired
	jz		popgtimer
	; spawn more enemies if time expired
	call		timeout_spawner

popgtimer:
	pop		dx
	call		checkwavetime

player_die:
	; If you uncomment the jmp instruction immediately below,
	; the player will be granted total invincibility.
	;jmp		win_level
	; Player Die ?
	test		[gflags],1
	jz		win_level
	call		death_animation
	dec		[lives]
	jnz		restart_level
	call		gameover
	jmp		wait_anykey

	; Wave Win ?
win_level:
	cmp		[wave],LAST_WAVE	; Wave 25 is never won!!
	jz		delay
	test		[gflags],2
	jz		delay
	; Do stuff for level win
	mov		ax,WAVE_BONUS
	mov		bl,[wave]
	xor		bh,bh
	mul		bx
	call		addpoints
	call		wavemsg
	inc		[wave]
restart_level:
	; Safety measure. Under normal operations, the last wave
	; should not end.
	cmp		[wave],LAST_WAVE
	jna		restart_level2
	mov		[wave],LAST_WAVE
restart_level2:
	call		init_level
delay:
	; After everything is done, wait until the next tick
	test		[otherkeys],1
	jnz		exit

	cmp		dx,[gtimer]
	je		delay
	jmp		main

; **** End of main loop ****

; -- EXIT PROGRAM ROUTINE --
; NOTE: Do not let this stray far from the other labels!!
exit:
	call		sound_kill

	; Reset tick rate
	mov		al,36h
	out		43h,al
	xor		ax,ax
	out		40h,al
	out		40h,al

	; Restore original system timer interrupt vector
	push		ds
	mov		ax,[oldpits]
	mov		dx,[oldpit]
	mov		ds,ax
	mov		ax,2508h
	int		21h
	pop		ds

	; Restore original interrupt 9 (keyboard) vector
	push		ds
	mov		ax,[oldint9s]
	mov		dx,[oldint9]
	mov		ds,ax
	mov		ax,2509h
	int		21h
	pop		ds

	; Set to 80x25 text color
	mov		ax,3
	int		10h
	
	; Shilling
	mov		ah,9
	mov		dx,exitstr
	int		21h
	
	mov		ax,4C00h
	int		21h
; -- END OF EXIT PROGRAM ROUTINE --

; -- PAUSE ROUTINE --
pause_loop:
; Called when player presses P to pause the game.
	call		sound_kill
	mov		dx,[gtimer]
pause_loop2:
	; Wait for the key to be released.
	test		[otherkeys],4
	jnz		pause_loop2
pause_loop3:
	call		sound_toggle_eval
	; Wait for it to be pressed again, or ESC to exit the program.
	test		[otherkeys],1
	jnz		pause_loopend
	test		[otherkeys],4
	jz		pause_loop3

pause_loopend:
	; Wait for the P key to be released again.
	test		[otherkeys],4
	jnz		pause_loopend
	mov		[gtimer],dx
	ret
; -- END OF PAUSE ROUTINE --

; -- INCLUDES --
; Yes, it matters that they are located here.
include 'extprocs.asm'
include 'sndprocs.asm'
include 'newints.asm'
