; Routines specifically dealing with the PC speaker.

soundstat	db	1

sweep_sound:
; A little bumpy sweep.
; Call with:
;	AX = PIT_FREQ/frequency
;	CX = sweep duration
;	DX = sweep step
; Preserves: CX
; Destroys: BX
	mov		bx,[gtimer]
sweep_sound_loop:
	call		sound_setfreq
	add		ax,dx
	call		gtimer_wait
	loop		sweep_sound_loop
	ret

SSHOOT		= 1
SESHOOT		= 2
SEDEATH		= 4
SEDASH		= 8
SBOUNCE		= 16
SBOMB		= 32
STIME		= 64

SSHOOT_TIME	= 5
SSHOOT_FREQ	= PIT_FREQ/3000
SEDEATH_TIME	= 4
SEDEATH_FREQ	= PIT_FREQ/1200
SEDASH_TIME	= 14
SEDASH_FREQ	= PIT_FREQ/4000
SBOUNCE_TIME	= 8
SBOUNCE_FREQ	= PIT_FREQ/200
SBOMB_TIME	= 20
SBOMB_FREQ	= PIT_FREQ/6000
STIME_TIME	= 30
STIME_FREQ	= PIT_FREQ/675

sound_setfreq:
; Routine to open up the PC speaker to start playing a new tone.
; Call with:
;	AX = PIT_FREQ/frequency
; Preserves: AX
	test		[soundstat],1
	jz		sound_abort
	push		ax
	mov		al,0B6h
	out		43h,al
	pop		ax
	push		ax
	out		42h,al
	mov		al,ah
	out		42h,al
	in		al,61h
	or		al,3
	out		61h,al
	pop		ax
sound_abort:
	; If the game is muted, do nothing
	ret

sound_loop:
; Persist sounds in between ticks.
; A higher bit sound specified in [sflags] takes priority.
;
; [stimer] counts in game ticks.

stime_start:
	test		[sflags],STIME
	jz		sbomb_start
	cmp		[stimer],0
	jnz		stime_continue
	mov		[stimer],STIME_TIME
	mov		[sfreq],STIME_FREQ

stime_continue:
	; The idea is to have two separate beeps.
	mov		al,[stimer]
	cmp		al,20
	jae		stime_continue2
	cmp		al,10
	jbe		stime_continue2
	jmp		stime_silent
stime_continue2:
	mov		ax,[sfreq]
	call		sound_setfreq

	dec		[stimer]
	jmp		sloop_end

sbomb_start:
	test		[sflags],SBOMB
	jz		sbounce_start
	cmp		[stimer],0
	jnz		sbomb_continue
	mov		[stimer],SBOMB_TIME
	mov		[sfreq],SBOMB_FREQ

sbomb_continue:
	mov		ax,[sfreq]
	call		sound_setfreq

	cmp		[stimer],SBOMB_TIME
	jne		sbomb_continue2
	mov		[sfreq],PIT_FREQ/500
	dec		[stimer]
	jmp		sloop_end
sbomb_continue2:
	add		[sfreq],800
	dec		[stimer]
	jmp		sloop_end

sedash_start:
	test		[sflags],SEDASH
	jz		sedeath_start
	cmp		[stimer],0
	jnz		sedash_continue
	mov		[stimer],SEDASH_TIME
	mov		[sfreq],SEDASH_FREQ
sedash_continue:
	mov		ax,[sfreq]
	call		sound_setfreq

	add		[sfreq],1000
	dec		[stimer]
	jmp		sloop_end

sbounce_start:
	test		[sflags],SBOUNCE
	jz		sedash_start
	cmp		[stimer],0
	jnz		sbounce_continue
	mov		[stimer],SBOUNCE_TIME
	mov		[sfreq],SBOUNCE_FREQ
sbounce_continue:
	mov		ax,[sfreq]
	call		sound_setfreq

	sub		[sfreq],1200
	dec		[stimer]
	jmp		sloop_end

sedeath_start:
	test		[sflags],SEDEATH
	jz		seshoot_start
	cmp		[stimer],0
	jnz		sedeath_continue
	mov		[stimer],SEDEATH_TIME
	mov		[sfreq],SEDEATH_FREQ
sedeath_continue:
	mov		ax,[sfreq]
	call		sound_setfreq

	; Very different frequencies each tick
	mov		al,[stimer]

	cmp		al,4
	jb		sedeath_3
	mov		word [sfreq],PIT_FREQ/750
	jmp		sedeath_end
sedeath_3:
	cmp		al,3
	jb		sedeath_2
	mov		word [sfreq],PIT_FREQ/350
	jmp		sedeath_end
sedeath_2:
	cmp		al,2
	jb		sedeath_1
	mov		word [sfreq],PIT_FREQ/600
	jmp		sedeath_end
sedeath_1:
	mov		word [sfreq],PIT_FREQ/400
sedeath_end:
	dec		[stimer]
	jmp		sloop_end

seshoot_start:
	test		[sflags],SESHOOT
	jz		sshoot_start
	cmp		[stimer],0
	jnz		seshoot_continue
	mov		[stimer],SSHOOT_TIME
	mov		[sfreq],SSHOOT_FREQ+2000
seshoot_continue:
	mov		ax,[sfreq]
	call		sound_setfreq

	add		[sfreq],800
	dec		[stimer]
	jmp		sloop_end

sshoot_start:
	test		[sflags],SSHOOT
	jz		sloop_end2
	cmp		[stimer],0
	jnz		sshoot_continue
	mov		[stimer],SSHOOT_TIME
	mov		[sfreq],SSHOOT_FREQ
sshoot_continue:
	; Open up PIT for modifying the sound
	mov		ax,[sfreq]
	call		sound_setfreq

	; Modify the frequency for the next tick
	add		[sfreq],600
	dec		[stimer]
	jmp		sloop_end

stime_silent:
	; Don't kill the sound entirely, the PC speaker is
	; just turned off at this time.
	mov		al,0B6h
	out		43h,al
	in		al,61h
	and		al,not 3
	out		61h,al
	dec		[stimer]
	jmp		sloop_end2

sloop_end:
	jnz		sloop_end2
	call		sound_kill
sloop_end2:
	ret

sound_kill:
; Stop playing sound.
; Destroys: AL
	mov		al,0B6h
	out		43h,al
	in		al,61h
	and		al,not 3
	out		61h,al
	mov		[sflags],0
	mov		[stimer],0
	mov		[sfreq],0
	ret

sound_toggle_eval:
	; Mute key hit?
	test		[otherkeys],8
	jz		mute_key_released
	; Gotta wait for it to be released before evaluating again
	test		[soundstat],2
	jnz		sound_toggle_end
	xor		[soundstat],1
	or		[soundstat],2
	jmp		sound_toggle_end

mute_key_released:
	and		[soundstat],not 2

sound_toggle_end:
	ret
